@@ -19,11 +19,37 @@ def print_memory_use():
     print("MEMORY",process.memory_info().rss/1000000000)
 
 
+def get_seqlet_neighbors_with_initcluster(
+        nearest_neighbors_to_compute,
+        coarse_affmat, initclusters): 
+    if (initclusters is not None):
+        assert len(initclusters)==len(coarse_affmat)
+    #get the argsort for coarse_affmat
+    coarse_affmat_argsort = np.argsort(-coarse_affmat, axis=-1)
+    nearest_neighbors = []
+    for row_idx,argsort_row in enumerate(coarse_affmat_argsort):
+        combined_neighbor_row = []
+        neighbor_row_topnn = argsort_row[:nearest_neighbors_to_compute+1]
+        neighbor_set_topnn = set(neighbor_row_topnn)
+        combined_neighbor_row.extend(neighbor_row_topnn)
+        if (initclusters is not None):
+            combined_neighbor_row.extend([
+                y for y in ([x for x in argsort_row
+                             if initclusters[x]==initclusters[row_idx]][
+                             :nearest_neighbors_to_compute+1])
+                if y not in neighbor_set_topnn])
+        nearest_neighbors.append(combined_neighbor_row) 
+    return nearest_neighbors
+
+
 class TfModiscoSeqletsToPatternsFactory(object):
 
     def __init__(self, n_cores=4,
                        min_overlap_while_sliding=0.7,
 
+                       #init clusterer factory
+                       initclusterer_factory=None,                       
+
                        #gapped kmer embedding arguments
                        alphabet_size=4,
                        kmer_len=8, num_gaps=3, num_mismatches=2,
@@ -37,10 +63,14 @@ class TfModiscoSeqletsToPatternsFactory(object):
                        skip_fine_grained=False,
 
                        tsne_perplexity = 10,
+                       use_louvain=False,
+                       louvain_initclusters_weight=1.0,
+                       n_leiden_iterations_r1=-1,
+                       n_leiden_iterations_r2=-1,
                        louvain_num_runs_and_levels_r1=[(200,-1)],
-                       louvain_num_runs_and_levels_r2=[(200,-1)],
-                       louvain_contin_runs_r1 = 50,
-                       louvain_contin_runs_r2 = 50,
+                       louvain_num_runs_and_levels_r2=[(200,-1)], 
+                       contin_runs_r1 = 50,
+                       contin_runs_r2 = 50,
                        final_louvain_level_to_return=1,
 
                        frac_support_to_trim_to=0.2,
@@ -63,6 +93,12 @@ class TfModiscoSeqletsToPatternsFactory(object):
                        final_flank_to_add=10,
                        verbose=True, seed=1234):
 
+        self.initclusterer_factory = initclusterer_factory
+        if (use_louvain==True):
+            assert self.initclusterer_factory==None,\
+                    ("Louvain doesn't support cluster initialization;"
+                     +" set use_louvain to False")
+
         #affinity_mat calculation
         self.n_cores = n_cores
         self.min_overlap_while_sliding = min_overlap_while_sliding
@@ -85,10 +121,14 @@ class TfModiscoSeqletsToPatternsFactory(object):
         self.tsne_perplexity = tsne_perplexity
 
         #clustering settings
+        self.use_louvain = use_louvain
+        self.louvain_initclusters_weight = louvain_initclusters_weight
+        self.n_leiden_iterations_r1 = n_leiden_iterations_r1
+        self.n_leiden_iterations_r2 = n_leiden_iterations_r2
+        self.contin_runs_r1 = contin_runs_r1
+        self.contin_runs_r2 = contin_runs_r2
         self.louvain_num_runs_and_levels_r1 = louvain_num_runs_and_levels_r1
         self.louvain_num_runs_and_levels_r2 = louvain_num_runs_and_levels_r2
-        self.louvain_contin_runs_r1 = louvain_contin_runs_r1
-        self.louvain_contin_runs_r2 = louvain_contin_runs_r2
         self.final_louvain_level_to_return = final_louvain_level_to_return
 
         #postprocessor1 settings
@@ -122,6 +162,8 @@ class TfModiscoSeqletsToPatternsFactory(object):
         to_return =  OrderedDict([
                 ('class_name', type(self).__name__),
                 ('n_cores', self.n_cores),
+                ('initclusterer_factory',
+                 self.initclusterer_factory.get_jsonable_config())
                 ('min_overlap_while_sliding', self.min_overlap_while_sliding),
                 ('alphabet_size', self.alphabet_size),
                 ('kmer_len', self.kmer_len),
@@ -134,16 +176,17 @@ class TfModiscoSeqletsToPatternsFactory(object):
                  self.affmat_correlation_threshold),
                 ('filter_beyond_first_round', filter_beyond_first_round),
                 ('tsne_perplexity', self.tsne_perplexity),
+                ('use_louvain', self.use_louvain),
                 ('louvain_num_runs_and_levels_r1',
                  self.louvain_num_runs_and_levels_r1),
                 ('louvain_num_runs_and_levels_r2',
                  self.louvain_num_runs_and_levels_r2),
                 ('final_louvain_level_to_return',
                  self.final_louvain_level_to_return),
-                ('louvain_contin_runs_r1',
-                 self.louvain_contin_runs_r1),
-                ('louvain_contin_runs_r2',
-                 self.louvain_contin_runs_r2),
+                ('contin_runs_r1',
+                 self.contin_runs_r1),
+                ('contin_runs_r2',
+                 self.contin_runs_r2),
                 ('frac_support_to_trim_to', self.frac_support_to_trim_to),
                 ('min_num_to_trim_to', self.min_num_to_trim_to),
                 ('trim_to_window_size', self.trim_to_window_size),
@@ -172,9 +215,13 @@ class TfModiscoSeqletsToPatternsFactory(object):
         seqlets_sorter = (lambda arr:
                           sorted(arr,
                                  key=lambda x:
-                                      -np.sum([np.sum(np.abs(x[track_name].fwd))
-                                         for track_name
-                                         in contrib_scores_track_names])))
+                                  -np.sum([np.sum(np.abs(x[track_name].fwd))
+                                     for track_name
+                                     in contrib_scores_track_names])))
+
+        if (self.initclusterer_factory is not None):
+            self.initclusterer_factory.set_onehot_track_name(onehot_track_name)
+        initclusterer_factory = self.initclusterer_factory
 
         pattern_comparison_settings =\
             affinitymat.core.PatternComparisonSettings(
@@ -208,10 +255,6 @@ class TfModiscoSeqletsToPatternsFactory(object):
                         gpu_batch_size=None),
                 verbose=self.verbose)
 
-        nearest_neighbors_computer = nearest_neighbors.ScikitNearestNeighbors(
-            n_neighbors=self.nearest_neighbors_to_compute,
-            nn_n_jobs=self.nn_n_jobs)  
-
         affmat_from_seqlets_with_nn_pairs =\
             affinitymat.core.AffmatFromSeqletsWithNNpairs(
                 pattern_comparison_settings=pattern_comparison_settings,
@@ -236,31 +279,46 @@ class TfModiscoSeqletsToPatternsFactory(object):
         affmat_transformer_r1 = affinitymat.transformers.SymmetrizeByAddition(
                                 probability_normalize=True)
         print("TfModiscoSeqletsToPatternsFactory: seed=%d" % self.seed)
-        for n_runs, level_to_return in self.louvain_num_runs_and_levels_r1:
-            affmat_transformer_r1 = affmat_transformer_r1.chain(
-                affinitymat.transformers.LouvainMembershipAverage(
-                    n_runs=n_runs,
-                    level_to_return=level_to_return,
-                    parallel_threads=self.n_cores, seed=self.seed))
-        clusterer_r1 = cluster.core.LouvainCluster(
-            level_to_return=self.final_louvain_level_to_return,
-            affmat_transformer=affmat_transformer_r1,
-            contin_runs=self.louvain_contin_runs_r1,
-            verbose=self.verbose, seed=self.seed)
+        if (self.use_louvain):
+            for n_runs, level_to_return in self.louvain_num_runs_and_levels_r1:
+                affmat_transformer_r1 = affmat_transformer_r1.chain(
+                    affinitymat.transformers.LouvainMembershipAverage(
+                        n_runs=n_runs,
+                        level_to_return=level_to_return,
+                        parallel_threads=self.n_cores, seed=self.seed))
+            clusterer_r1 = cluster.core.LouvainCluster(
+                level_to_return=self.final_louvain_level_to_return,
+                affmat_transformer=affmat_transformer_r1,
+                contin_runs=self.contin_runs_r1,
+                verbose=self.verbose, seed=self.seed)
+        else:
+            clusterer_r1 = cluster.core.LeidenCluster(
+                affmat_transformer=affmat_transformer_r1,
+                contin_runs=self.contin_runs_r1,
+                n_leiden_iterations=self.n_leiden_iterations_r1,
+                verbose=self.verbose)
 
         affmat_transformer_r2 = affinitymat.transformers.SymmetrizeByAddition(
                                 probability_normalize=True)
-        for n_runs, level_to_return in self.louvain_num_runs_and_levels_r2:
-            affmat_transformer_r2 = affmat_transformer_r2.chain(
-                affinitymat.transformers.LouvainMembershipAverage(
-                    n_runs=n_runs,
-                    level_to_return=level_to_return,
-                    parallel_threads=self.n_cores, seed=self.seed))
-        clusterer_r2 = cluster.core.LouvainCluster(
-            level_to_return=self.final_louvain_level_to_return,
-            affmat_transformer=affmat_transformer_r2,
-            contin_runs=self.louvain_contin_runs_r2,
-            verbose=self.verbose, seed=self.seed)
+        if (self.use_louvain):
+            for n_runs, level_to_return in self.louvain_num_runs_and_levels_r2:
+                affmat_transformer_r2 = affmat_transformer_r2.chain(
+                    affinitymat.transformers.LouvainMembershipAverage(
+                        n_runs=n_runs,
+                        level_to_return=level_to_return,
+                        parallel_threads=self.n_cores, seed=self.seed))
+            clusterer_r2 = cluster.core.LouvainCluster(
+                level_to_return=self.final_louvain_level_to_return,
+                affmat_transformer=affmat_transformer_r2,
+                contin_runs=self.contin_runs_r2,
+                verbose=self.verbose, seed=self.seed,
+                initclusters_weight=self.louvain_initclusters_weight)
+        else:
+            clusterer_r2 = cluster.core.LeidenCluster(
+                affmat_transformer=affmat_transformer_r2,
+                contin_runs=self.contin_runs_r2,
+                n_leiden_iterations=self.n_leiden_iterations_r2,
+                verbose=self.verbose)
         
         clusterer_per_round = [clusterer_r1, clusterer_r2]
 
@@ -396,8 +454,9 @@ class TfModiscoSeqletsToPatternsFactory(object):
 
         return TfModiscoSeqletsToPatterns(
                 seqlets_sorter=seqlets_sorter,
+                initclusterer_factory=initclusterer_factory,
                 coarse_affmat_computer=coarse_affmat_computer,
-                nearest_neighbors_computer=nearest_neighbors_computer,
+                nearest_neighbors_to_compute=self.nearest_neighbors_to_compute,
                 affmat_from_seqlets_with_nn_pairs=
                     affmat_from_seqlets_with_nn_pairs, 
                 filter_mask_from_correlation=filter_mask_from_correlation,
@@ -422,23 +481,60 @@ class TfModiscoSeqletsToPatternsFactory(object):
 class SeqletsToPatternsResults(object):
 
     def __init__(self,
+                 each_round_initcluster_motifs, 
                  patterns, cluster_results,
                  total_time_taken, success=True, **kwargs):
+        self.each_round_initcluster_motifs = each_round_initcluster_motifs
         self.success = success
         self.patterns = patterns
         self.cluster_results = cluster_results
         self.total_time_taken = total_time_taken
         self.__dict__.update(**kwargs)
 
+    def save_each_round_initcluster_motifs(self, grp):
+        all_round_names = []
+        for (round_idx,initcluster_motifs)\
+            in enumerate(self.each_round_initcluster_motifs):
+            round_name = "round_"+str(round_idx)
+            util.save_patterns(patterns=initcluster_motifs,
+                               grp=grp.create_group(round_name)) 
+        util.save_string_list(
+            string_list=all_round_names,
+            dset_name="all_round_names",          
+            grp=grp)
+
+    @classmethod
+    def load_each_round_initcluster_motifs(cls, grp, track_set):
+        all_round_names = util.load_string_list(dset_name="all_round_names",
+                                                grp=grp) 
+        each_round_initcluster_motifs = [] 
+        for round_name in all_round_names:
+            round_grp = grp[round_name]
+            initcluster_motifs = load_patterns(grp=round_grp,
+                                               track_set=track_set)
+            each_round_initcluster_motifs.append(initcluster_motifs)
+        return each_round_initcluster_motifs
+             
     @classmethod
     def from_hdf5(cls, grp, track_set):
         success = grp.attrs.get("success", False)
         if (success):
-            patterns = util.load_patterns(grp=grp["patterns"], track_set=track_set) 
+            if ("each_round_initcluster_motifs" not in grp):
+                each_round_initcluster_motifs = None 
+            else:
+                each_round_initcluster_motifs =\
+                    cls.load_each_round_initcluster_motifs(
+                        grp=grp["each_round_initcluster_motifs"],
+                        track_set=track_set)
+            patterns = util.load_patterns(grp=grp["patterns"],
+                                          track_set=track_set) 
             cluster_results = None
             total_time_taken = None
-            return cls(patterns=patterns, cluster_results=cluster_results,
-                       total_time_taken=total_time_taken)
+            return cls(
+                each_round_initcluster_motifs=each_round_initcluster_motifs,
+                patterns=patterns,
+                cluster_results=cluster_results,
+                total_time_taken=total_time_taken)
         else:
             return cls(success=False, patterns=None, cluster_results=None,
                        total_time_taken=None)
@@ -446,6 +542,9 @@ class SeqletsToPatternsResults(object):
     def save_hdf5(self, grp):
         grp.attrs["success"] = self.success
         if (self.success):
+            if (self.each_round_initcluster_motifs is not None):
+                self.save_each_round_initcluster_motifs(
+                    grp=grp.create_group("each_round_initcluster_motifs"))
             util.save_patterns(self.patterns,
                                grp.create_group("patterns"))
             self.cluster_results.save_hdf5(grp.create_group("cluster_results"))   
@@ -461,8 +560,9 @@ class AbstractSeqletsToPatterns(object):
 class TfModiscoSeqletsToPatterns(AbstractSeqletsToPatterns):
 
     def __init__(self, seqlets_sorter, 
+                       initclusterer_factory,
                        coarse_affmat_computer,
-                       nearest_neighbors_computer,
+                       nearest_neighbors_to_compute,
                        affmat_from_seqlets_with_nn_pairs, 
                        filter_mask_from_correlation,
                        filter_beyond_first_round,
@@ -478,8 +578,9 @@ class TfModiscoSeqletsToPatterns(AbstractSeqletsToPatterns):
                        verbose=True):
 
         self.seqlets_sorter = seqlets_sorter
+        self.initclusterer_factory = initclusterer_factory
         self.coarse_affmat_computer = coarse_affmat_computer
-        self.nearest_neighbors_computer = nearest_neighbors_computer
+        self.nearest_neighbors_to_compute = nearest_neighbors_to_compute
         self.affmat_from_seqlets_with_nn_pairs =\
             affmat_from_seqlets_with_nn_pairs
         self.filter_mask_from_correlation = filter_mask_from_correlation
@@ -499,9 +600,48 @@ class TfModiscoSeqletsToPatterns(AbstractSeqletsToPatterns):
         self.verbose = verbose
 
 
+    def get_cluster_to_aggregate_motif(self, seqlets, cluster_indices,
+                                       sign_consistency_check,
+                                       min_seqlets_in_motif):
+        num_clusters = max(cluster_indices+1)
+        cluster_to_seqlets = defaultdict(list) 
+        assert len(seqlets)==len(cluster_indices)
+        for seqlet,idx in zip(seqlets, cluster_indices):
+            cluster_to_seqlets[idx].append(seqlet)
+
+        cluster_to_motif = OrderedDict()
+        cluster_to_eliminated_motif = OrderedDict()
+        for i in range(num_clusters):
+            if (len(cluster_to_seqlets[i]) >= min_seqlets_in_motif):
+                if (self.verbose):
+                    print("Aggregating for cluster "+str(i)+" with "
+                          +str(len(cluster_to_seqlets[i]))+" seqlets")
+                    print_memory_use()
+                    sys.stdout.flush()
+                motifs = self.seqlet_aggregator(cluster_to_seqlets[i])
+                assert len(motifs)<=1
+                if (len(motifs) > 0):
+                    motif = motifs[0]
+                    if (sign_consistency_check==False or
+                        self.sign_consistency_func(motif)):
+                        cluster_to_motif[i] = motif
+                    else:
+                        if (self.verbose):
+                            print("Dropping cluster "+str(i)+
+                                  " with "+str(motif.num_seqlets)
+                                  +" seqlets due to sign disagreement")
+                        cluster_to_eliminated_motif[i] = motif
+                cluster_to_motif[i] = motif
+        return cluster_to_motif, cluster_to_eliminated_motif
+
+
     def __call__(self, seqlets):
 
         seqlets = self.seqlets_sorter(seqlets)
+        if (self.initclusterer_factory is not None):
+            initclusterer = self.initclusterer_factory(seqlets=seqlets) 
+        else:
+            initclusterer = None
 
         start = time.time()
 
@@ -515,13 +655,29 @@ class TfModiscoSeqletsToPatterns(AbstractSeqletsToPatterns):
         #cluster_to_motif_sets = []
         #cluster_to_eliminated_motif_sets = []
 
+        if (initclusterer is not None):
+            each_round_initcluster_motifs = []
+        else:
+            each_round_initcluster_motifs = None
+
         for round_idx, clusterer in enumerate(self.clusterer_per_round):
             import gc
             gc.collect()
 
             round_num = round_idx+1
 
-            #seqlets_sets.append(seqlets)
+            if (initclusterer is not None): 
+                initclusters = initclusterer(seqlets=seqlets)
+                initcluster_motifs =\
+                    list(self.get_cluster_to_aggregate_motif(
+                            seqlets=seqlets,
+                            cluster_indices=initclusters,
+                            sign_consistency_check=False,
+                            min_seqlets_in_motif=2)[0].values())
+                each_round_initcluster_motifs.append(initcluster_motifs)
+            else:
+                initclusters = None
+                initcluster_motifs = None
             
             if (len(seqlets)==0):
                 if (self.verbose):
@@ -551,7 +707,11 @@ class TfModiscoSeqletsToPatterns(AbstractSeqletsToPatterns):
                     print_memory_use()
                     sys.stdout.flush()
 
-                seqlet_neighbors = self.nearest_neighbors_computer(coarse_affmat)
+                seqlet_neighbors = get_seqlet_neighbors_with_initcluster(
+                    nearest_neighbors_to_compute=
+                     self.nearest_neighbors_to_compute,
+                    coarse_affmat=coarse_affmat,
+                    initclusters=initclusters)
 
                 if (self.verbose):
                     print("Computed nearest neighbors in",
@@ -566,8 +726,8 @@ class TfModiscoSeqletsToPatterns(AbstractSeqletsToPatterns):
                     print_memory_use()
                     sys.stdout.flush()
                 nn_affmat = self.affmat_from_seqlets_with_nn_pairs(
-                                            seqlet_neighbors=seqlet_neighbors,
-                                            seqlets=seqlets) 
+                                    seqlet_neighbors=seqlet_neighbors,
+                                    seqlets=seqlets) 
                 #nn_affmats.append(nn_affmat)
                 
                 if (self.verbose):
@@ -579,6 +739,8 @@ class TfModiscoSeqletsToPatterns(AbstractSeqletsToPatterns):
 
                 #filter by correlation
                 if (round_idx == 0 or self.filter_beyond_first_round==True):
+                    #the filter_mask_from_correlation function only operates
+                    # on columns in which np.abs(main_affmat) > 0
                     filtered_rows_mask = self.filter_mask_from_correlation(
                                             main_affmat=nn_affmat,
                                             other_affmat=coarse_affmat) 
@@ -598,7 +760,11 @@ class TfModiscoSeqletsToPatterns(AbstractSeqletsToPatterns):
                         sys.stdout.flush()
 
                 filtered_seqlets = [x[0] for x in
-                           zip(seqlets, filtered_rows_mask) if (x[1])]
+                    zip(seqlets, filtered_rows_mask) if (x[1])]
+                if (initclusters is not None):
+                    filtered_initclusters = initclusters[filtered_rows_mask] 
+                else:
+                    filtered_initclusters = None
                 #filtered_seqlets_sets.append(filtered_seqlets)
 
                 filtered_affmat =\
@@ -608,6 +774,10 @@ class TfModiscoSeqletsToPatterns(AbstractSeqletsToPatterns):
             else:
                 filtered_affmat = coarse_affmat
                 filtered_seqlets = seqlets
+                if (initclusters is not None):
+                    filtered_initclusters = initclusters
+                else:
+                    filtered_initclusters = None
 
             if (self.verbose):
                 print("(Round "+str(round_num)+") Computing density "
@@ -625,7 +795,8 @@ class TfModiscoSeqletsToPatterns(AbstractSeqletsToPatterns):
                 print_memory_use()
                 sys.stdout.flush() 
 
-            cluster_results = clusterer(density_adapted_affmat)
+            cluster_results = clusterer(density_adapted_affmat,
+                                        initclusters=filtered_initclusters)
             del density_adapted_affmat
             #cluster_results_sets.append(cluster_results)
             num_clusters = max(cluster_results.cluster_indices+1)
@@ -644,35 +815,12 @@ class TfModiscoSeqletsToPatterns(AbstractSeqletsToPatterns):
                 print_memory_use()
                 sys.stdout.flush()
 
-            cluster_to_seqlets = defaultdict(list) 
-            assert len(filtered_seqlets)==len(cluster_results.cluster_indices)
-            for seqlet,idx in zip(filtered_seqlets,
-                                  cluster_results.cluster_indices):
-                cluster_to_seqlets[idx].append(seqlet)
-
-            cluster_to_eliminated_motif = OrderedDict()
-            cluster_to_motif = OrderedDict()
-            #cluster_to_motif_sets.append(cluster_to_motif)
-            #cluster_to_eliminated_motif_sets.append(
-            #    cluster_to_eliminated_motif)
-            for i in range(num_clusters):
-                if (self.verbose):
-                    print("Aggregating for cluster "+str(i)+" with "
-                          +str(len(cluster_to_seqlets[i]))+" seqlets")
-                    print_memory_use()
-                    sys.stdout.flush()
-                motifs = self.seqlet_aggregator(cluster_to_seqlets[i])
-                assert len(motifs)<=1
-                if (len(motifs) > 0):
-                    motif = motifs[0]
-                    if (self.sign_consistency_func(motif)):
-                        cluster_to_motif[i] = motif
-                    else:
-                        if (self.verbose):
-                            print("Dropping cluster "+str(i)+
-                                  " with "+str(motif.num_seqlets)
-                                  +" seqlets due to sign disagreement")
-                        cluster_to_eliminated_motif[i] = motif
+            cluster_to_motif, cluster_to_eliminated_motif =\
+                self.get_cluster_to_aggregate_motif(
+                    seqlets=filtered_seqlets,
+                    cluster_indices=cluster_results.cluster_indices,
+                    sign_consistency_check=True,
+                    min_seqlets_in_motif=0)
 
             #obtain unique seqlets from adjusted motifs
             seqlets = dict([(y.exidx_start_end_string, y)
@@ -733,6 +881,7 @@ class TfModiscoSeqletsToPatterns(AbstractSeqletsToPatterns):
             sys.stdout.flush()
 
         results = SeqletsToPatternsResults(
+            each_round_initcluster_motifs=each_round_initcluster_motifs,             
             patterns=final_patterns,
             seqlets=filtered_seqlets, #last stage of filtered seqlets
             #affmat=filtered_affmat,
